import pytest

from dipin import Container
from dipin.container import PartialFactoryContainerItem


class A: ...


def test_register_single_factory():
    container = Container()

    container.register_factory(A)

    assert len(container) == 1
    assert (A, None) in container

    item = container.get((A, None))
    assert isinstance(item, PartialFactoryContainerItem)

    factory_fn = item.factory
    assert factory_fn is A
    assert isinstance(factory_fn(), A)

    assert item.use_cache is False


def test_factory_function_returns_different_instances():
    container = Container()

    container.register_factory(A)

    prev_instance_id = 0
    for i in range(3):
        assert id(container[(A, None)].factory()) != prev_instance_id


def test_registering_two_factories_of_same_type_replaces_instance_and_emits_warning():
    class B:
        def __init__(self, val: int):
            self.val = val

    container = Container()

    container.register_factory(B, lambda: B(1))

    with pytest.warns() as record:
        container.register_factory(B, lambda: B(2))

    assert len(record) == 1
    assert (
        str(record[0].message)
        == f"Replacing existing container item {B.__module__}.{B.__qualname__}"
    )

    assert len(container) == 1
    assert (B, None) in container

    factory_fn = container[(B, None)].factory
    assert callable(factory_fn)
    assert factory_fn().val == 2



import pytest

from dipin import Container
from dipin.container import InstanceContainerItem


class A: ...


class B: ...


def test_initial_container_size():
    container = Container()

    assert len(container) == 0


def test_register_single_instance():
    container = Container()

    instance = A()
    container.register_instance(instance)

    assert len(container) == 1
    assert (A, None) in container

    item = container[(A, None)]
    assert isinstance(item, InstanceContainerItem)

    assert item.instance is instance


def test_registering_two_types_of_instance():
    container = Container()

    instance_a = A()
    instance_b = B()

    container.register_instance(instance_a)
    container.register_instance(instance_b)

    assert len(container) == 2
    assert (A, None) in container
    assert (B, None) in container

    item_a = container[(A, None)]
    assert isinstance(item_a, InstanceContainerItem)
    assert item_a.instance is instance_a

    item_b = container[(B, None)]
    assert isinstance(item_b, InstanceContainerItem)
    assert item_b.instance is instance_b


def test_registering_two_instances_of_same_type_replaces_instance_and_emits_warning():
    container = Container()

    instance_a = A()
    instance_b = A()

    container.register_instance(instance_a)

    item = container[(A, None)]
    assert item.instance is instance_a

    with pytest.warns() as record:
        container.register_instance(instance_b)

    assert len(record) == 1
    assert (
        str(record[0].message)
        == f"Replacing existing container item {A.__module__}.{A.__qualname__}"
    )

    assert len(container) == 1
    assert (A, None) in container

    item = container[(A, None)]
    assert item.instance is instance_b



__version__ = "0.0.3"

from .container import Container
from .interface import FastAPIContainer
from .resolver import CircularDependencyError

__all__ = ["DI", "Container", "CircularDependencyError"]


DI = FastAPIContainer()



from dipin import Container
from dipin.interface import ResolvingContainer
from dipin.resolver import Resolver


def test_resolving_container_caches_create_once_factories():
    class A: ...

    container = Container()
    container.register_factory(A, create_once=True)

    item = container[(A, None)]
    assert item.use_cache is True

    resolver = Resolver(container)
    interface = ResolvingContainer(container, resolver)

    res = interface.retrieve((A, None))
    assert isinstance(res, A)

    res_2 = interface.retrieve((A, None))
    assert res is res_2


def test_resolving_container_doesnt_cache_other_factories():
    class A: ...

    container = Container()
    container.register_factory(A)

    item = container[(A, None)]
    assert item.use_cache is False

    resolver = Resolver(container)
    interface = ResolvingContainer(container, resolver)

    res = interface.retrieve((A, None))
    assert isinstance(res, A)

    res_2 = interface.retrieve((A, None))
    assert res is not res_2



import pytest

from dipin.interface import ResolvingContainer
from dipin.resolver import UnfillableArgumentError


def test_autowiring_of_unregistered_dependencies():
    DI = ResolvingContainer()

    class Service: ...

    assert len(DI) == 0
    assert Service not in DI

    svc = DI.get(Service)
    assert isinstance(svc, Service)

    assert len(DI) == 1
    assert Service in DI


def test_autowiring_of_unregistered_deep_dependencies():
    DI = ResolvingContainer()

    class Service: ...

    class DependentService:
        def __init__(self, svc: Service):
            self.svc = svc

    assert len(DI) == 0
    assert Service not in DI
    assert DependentService not in DI

    svc = DI.get(DependentService)
    assert isinstance(svc, DependentService)

    assert len(DI) == 2
    assert Service in DI
    assert DependentService in DI


def test_autowiring_of_unregistered_dependency_with_non_type_arg():
    DI = ResolvingContainer()

    class Service:
        def __init__(self, api_token: str): ...

    assert len(DI) == 0
    assert Service not in DI

    with pytest.raises(UnfillableArgumentError) as e:
        DI.get(Service)

    assert str(e.value) == "Unable to fill parameter api_token (<class 'str'>)"



from typing import Annotated, get_origin, get_args
from fastapi import FastAPI, Depends, params
from fastapi.testclient import TestClient

from dipin.interface import FastAPIContainer


def test_fastapi_depends_accessor():
    DI = FastAPIContainer()

    class Service: ...

    DI.register_factory(Service)

    dependency = DI[Service]

    origin = get_origin(dependency)
    assert origin is Annotated

    real_type, depends = get_args(dependency)
    assert real_type is Service
    assert isinstance(depends, params.Depends)
    assert depends.use_cache is False

    result = dependency()
    assert isinstance(result, Service)


def test_fastapi_dependency_caching_is_not_used():
    DI = FastAPIContainer()

    class Service: ...

    DI.register_factory(Service, create_once=True)

    dependency = DI[Service]
    _, depends = get_args(dependency)
    assert depends.use_cache is False

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[Service]) -> int:
        return id(svc)

    test_client = TestClient(app)
    expected_instance_id = id(DI.get(Service))
    for _ in range(3):
        resp = test_client.get("/")
        instance_id = int(resp.text)

        assert instance_id == expected_instance_id


def test_stock_fastapi_experience():
    class Service: ...

    class DependentService:
        service: Service

        def __init__(self, service: Annotated[Service, Depends()]):
            self.service = service

    app = FastAPI()

    @app.get("/")
    async def test(svc: Annotated[DependentService, Depends()]):
        return {
            "svc_type": type(svc).__name__,
            "svc_service_type": type(svc.service).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["svc_type"] == "DependentService"
        assert json["svc_service_type"] == "Service"


def test_di_fastapi_experience():
    DI = FastAPIContainer()

    class Service: ...

    DI.register_factory(Service)

    class DependentService:
        service: Service

        def __init__(self, service: DI[Service]):
            self.service = service

    DI.register_factory(DependentService)

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[DependentService]):
        return {
            "svc_type": type(svc).__name__,
            "svc_service_type": type(svc.service).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["svc_type"] == "DependentService"
        assert json["svc_service_type"] == "Service"


def test_fastapi_simple_instance_injection():
    DI = FastAPIContainer()

    class Service: ...

    instance = Service()
    DI.register_instance(instance)

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[Service]):
        return {
            "type": type(svc).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["type"] == "Service"
        assert json["id"] == id(instance)


def test_fastapi_simple_factory_function_injection():
    DI = FastAPIContainer()

    class Service: ...

    def factory() -> Service:
        return Service()

    DI.register_factory(Service, factory)

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[Service]):
        return {
            "type": type(svc).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    last_instance_id = None
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["type"] == "Service"
        if last_instance_id:
            assert json["id"] != last_instance_id
            last_instance_id = json["id"]


def test_fastapi_simple_factory_class_injection():
    DI = FastAPIContainer()

    class Service: ...

    DI.register_factory(Service)

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[Service]):
        return {
            "type": type(svc).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    last_instance_id = None
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["type"] == "Service"
        if last_instance_id:
            assert json["id"] != last_instance_id
            last_instance_id = json["id"]


def test_fastapi_dependent_services_factory():
    DI = FastAPIContainer()

    class Service: ...

    DI.register_factory(Service)

    class DependentService:
        service: Service

        def __init__(self, service: Service):
            self.service = service

    DI.register_factory(DependentService, lambda: DependentService(DI.get(Service)))

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[DependentService]):
        return {
            "svc_type": type(svc).__name__,
            "svc_service_type": type(svc.service).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["svc_type"] == "DependentService"
        assert json["svc_service_type"] == "Service"


def test_fastapi_dependent_services_resolved():
    DI = FastAPIContainer()

    class Service: ...

    DI.register_factory(Service)

    class DependentService:
        service: Service

        def __init__(self, service: DI[Service]):
            self.service = service

    DI.register_factory(DependentService)

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[DependentService]):
        return {
            "svc_type": type(svc).__name__,
            "svc_service_type": type(svc.service).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    for _ in range(3):
        resp = test_client.get("/")
        json = resp.json()

        assert json["svc_type"] == "DependentService"
        assert json["svc_service_type"] == "Service"


def test_fastapi_autowiring():
    DI = FastAPIContainer()

    class Service: ...

    class DependentService:
        service: Service

        def __init__(self, service: Service):
            self.service = service

    app = FastAPI()

    @app.get("/")
    async def test(svc: DI[DependentService]):
        return {
            "svc_type": type(svc).__name__,
            "svc_service_type": type(svc.service).__name__,
            "id": id(svc),
        }

    test_client = TestClient(app)
    resp = test_client.get("/")
    json = resp.json()

    assert json["svc_type"] == "DependentService"
    assert json["svc_service_type"] == "Service"



from typing import Type


def is_class_type(t: Type) -> bool:
    """This is hacky, but I don't know of a good method for checking if a type is a user-defined class."""

    return t not in (
        str,
        int,
        float,
        bool,
        list,
        dict,
        tuple,
        set,
        frozenset,
        type(None),
    )



from dipin.util import is_class_type


def test_class_detection():
    assert is_class_type(str) is False
    assert is_class_type(int) is False
    assert is_class_type(list) is False
    assert is_class_type(tuple) is False
    assert is_class_type(dict) is False
    assert is_class_type(set) is False
    assert is_class_type(bool) is False
    assert is_class_type(float) is False
    assert is_class_type(type(None)) is False

    class A: ...

    assert is_class_type(A) is True



import warnings
from dataclasses import dataclass
from typing import Callable, Type, TypeVar

from dipin.util import is_class_type

T = TypeVar("T")


InstanceType = Type[T]
Instance = T
Factory = Callable[[], T]
Name = str
LookupKey = InstanceType | Name
ContainerKey = tuple[InstanceType, Name | None]


@dataclass
class InstanceContainerItem:
    instance: Instance


@dataclass
class PartialFactoryContainerItem:
    use_cache: bool
    factory: Factory


@dataclass
class DefinedFactoryContainerItem:
    use_cache: bool
    factory: Factory


ContainerItem = (
    InstanceContainerItem | PartialFactoryContainerItem | DefinedFactoryContainerItem
)


class Container:
    container: dict[ContainerKey, ContainerItem]
    cache: dict[ContainerKey, Instance]

    def __init__(self):
        self.container = {}
        self.cache = {}

    def register_instance(
        self,
        instance: Instance,
        type_: InstanceType | None = None,
        name: Name | None = None,
    ) -> ContainerKey:
        if type_ is None:
            type_ = type(instance)

        if name:
            self._check_for_existing_names(name)

        self.set((type_, name), InstanceContainerItem(instance=instance))
        return type_, name

    def register_factory(
        self,
        type_: InstanceType,
        factory: Factory | None = None,
        name: Name | None = None,
        create_once: bool = False,
    ) -> ContainerKey:
        if name:
            self._check_for_existing_names(name)

        if factory is None:
            if not is_class_type(type_):
                raise ValueError(
                    "Omitting a factory function is only supported for classes"
                )
            self.set(
                (type_, name),
                PartialFactoryContainerItem(factory=type_, use_cache=create_once),
            )
            return type_, name

        self.set(
            (type_, name),
            DefinedFactoryContainerItem(factory=factory, use_cache=create_once),
        )
        return type_, name

    def set(self, key: ContainerKey, item: ContainerItem):
        if key in self.container:
            type_name = ".".join([key[0].__module__, key[0].__qualname__])
            name = f" (named '{key[1]}')" if key[1] else ""
            warnings.warn(
                UserWarning(f"Replacing existing container item {type_name}{name}")
            )

        self.container[key] = item

    def get(self, key: ContainerKey) -> ContainerItem:
        return self.container[key]

    def lookup(self, key: LookupKey) -> ContainerKey:
        if isinstance(key, Name):
            type_, name = self._find_by_name(key)
            return type_, name

        if (key, None) not in self.container:
            raise KeyError(f"Container item with type {key} not registered")

        return key, None

    def _find_by_name(self, name: str) -> ContainerKey:
        for type_, item_name in self.container.keys():
            if item_name == name:
                return type_, item_name

        raise KeyError(f"Container item with name {name} not registered")

    def _check_for_existing_names(self, name: str):
        try:
            self._find_by_name(name)
        except KeyError:
            # TODO: Handle replacements later
            raise KeyError(f"Existing container item with name {name}")

    def should_cache(self, key: ContainerKey) -> bool:
        item = self.container[key]

        if isinstance(item, InstanceContainerItem):
            return False

        return item.use_cache

    def is_cached(self, key: ContainerKey) -> bool:
        return key in self.cache

    def get_cached(self, key: ContainerKey) -> Instance:
        return self.cache[key]

    def set_cached(self, key: ContainerKey, instance: Instance):
        self.cache[key] = instance

    def __len__(self) -> int:
        return len(self.container)

    def __contains__(self, item: ContainerKey) -> bool:
        return item in self.container

    def __getitem__(self, item: ContainerKey) -> ContainerItem:
        return self.container[item]



from functools import partial
from typing import Annotated

from dipin.resolver import Resolver
from dipin.container import (
    Instance,
    ContainerKey,
    Container,
    InstanceType,
    Name,
    Factory,
    LookupKey,
)
from fastapi import Depends


class ResolvingContainer:
    """High-level interface for the DI container, using a dependency resolver"""

    container: Container
    resolver: Resolver

    def __init__(
        self, container: Container | None = None, resolver: Resolver | None = None
    ):
        self.container = container or Container()
        self.resolver = resolver or Resolver(self.container)

    def register_instance(self, instance: Instance, name: str | None = None) -> None:
        type_ = type(instance)

        self.container.register_instance(instance, type_, name)

    def register_factory(
        self,
        type_: InstanceType,
        factory: Factory | None = None,
        name: Name | None = None,
        create_once: bool = False,
    ) -> None:
        self.container.register_factory(type_, factory, name, create_once)

    def get(self, key: LookupKey) -> Instance:
        container_key = self.get_potential_key(key)
        return self.retrieve(container_key)

    def get_potential_key(self, key: LookupKey) -> ContainerKey:
        try:
            container_key = self.container.lookup(key)
        except KeyError as e:
            # Auto-wire type-lookups
            if not (container_key := self.resolver.autowire(key)):
                raise e

        return container_key

    def __getitem__(self, item: LookupKey) -> Instance:
        return self.get(item)

    def retrieve(self, container_key: ContainerKey) -> Instance:
        if self.container.is_cached(container_key):
            return self.container.get_cached(container_key)

        instance = self.resolver.get(container_key)

        if self.container.should_cache(container_key):
            self.container.set_cached(container_key, instance)

        return instance

    def __len__(self) -> int:
        return len(self.container)

    def __contains__(self, item: LookupKey) -> bool:
        try:
            self.container.lookup(item)
            return True
        except KeyError:
            return False


class FastAPIContainer(ResolvingContainer):
    """High-level interface for the DI container, for FastAPI application"""

    def __getitem__(self, key: LookupKey) -> Depends:
        container_key = self.get_potential_key(key)

        return Annotated[
            container_key[0],
            Depends(partial(self.retrieve, container_key), use_cache=False),
        ]



import inspect
from functools import partial
from typing import Generator, AsyncGenerator, get_args, Type

import asyncer
from dipin.container import (
    InstanceType,
    ContainerKey,
    Factory,
    Container,
    Instance,
    InstanceContainerItem,
    DefinedFactoryContainerItem,
    PartialFactoryContainerItem,
)
from dipin.util import is_class_type

DependencyTree = dict[ContainerKey, dict[str, ContainerKey]]


class Resolver:
    container: Container

    def __init__(self, container: Container):
        self.container = container

    def get(self, key: ContainerKey) -> Instance:
        # If the key is not in the container, attempt to autowire it
        if key not in self.container:
            try:
                if not (key_ := self.autowire(key[0])):
                    raise KeyError(f"Unable to resolve {key}")
            except UnfillableArgumentError as e:
                e.dependency = key
                raise e
            key = key_

        item = self.container.get(key)

        if isinstance(item, InstanceContainerItem):
            return item.instance

        try:
            assert isinstance(
                item, (DefinedFactoryContainerItem, PartialFactoryContainerItem)
            )
            factory = self.build_factory_from_factory(item.factory)
            return self.call_factory(factory)
        except RecursionError:
            # TODO: Detect this earlier
            raise CircularDependencyError(key)

    def call_factory(self, factory: Factory) -> Instance:
        if inspect.iscoroutinefunction(factory):
            return asyncer.syncify(factory)()

        result = factory()

        if isinstance(result, AsyncGenerator):

            async def get_next_item(g: AsyncGenerator) -> Instance:
                async for item in g:
                    return item

            return asyncer.syncify(get_next_item)(result)

        if isinstance(result, Generator):
            return next(result)

        return result

    def build_factory_from_factory(self, factory: Factory) -> Factory:
        return self.build_factory_dependencies(factory)

    def build_factory_dependencies(self, factory: Factory) -> Factory:
        args = inspect.signature(factory)

        params = {}
        for name, param in args.parameters.items():
            # Attempt to fetch/autowire dependencies
            if param.annotation is not inspect.Parameter.empty:
                if isinstance(param.annotation, str):
                    raise NotImplementedError(
                        "String annotations are not supported yet"
                    )

                if is_class_type(param.annotation):
                    anno_args = get_args(param.annotation)
                    if len(anno_args) > 0:
                        params[name] = self.get((anno_args[0], None))
                        continue

                    params[name] = self.get((param.annotation, None))
                    continue

            # Use default values
            if param.default is not inspect.Parameter.empty:
                params[name] = param.default
                continue

            raise UnfillableArgumentError(name, param.annotation)

        return partial(factory, **params)

    def autowire(self, type_: InstanceType) -> ContainerKey | None:
        if not self.can_autowire(type_):
            return None

        return self.container.register_factory(type_)

    def can_autowire(self, type_: InstanceType) -> bool:
        return is_class_type(type_)


class ResolverError(RuntimeError): ...


class UnfillableArgumentError(ResolverError):
    arg_name: str
    arg_type: Type
    dependency: ContainerKey | None

    def __init__(
        self, arg_name: str, arg_type: Type, dependency: ContainerKey | None = None
    ):
        self.arg_name = arg_name
        self.arg_type = arg_type
        self.dependency = dependency

    def __str__(self) -> str:
        return f"Unable to fill parameter {self.arg_name} ({self.arg_type})" + (
            " for {self.dependency}" if self.dependency else ""
        )


class CircularDependencyError(ResolverError):
    dependency: ContainerKey

    def __init__(self, dependency: ContainerKey):
        self.dependency = dependency

    def __str__(self) -> str:
        return f"Cannot construct {self.dependency} due to a circular dependency"



import pytest

from dipin.container import Container
from dipin.resolver import Resolver, CircularDependencyError, UnfillableArgumentError


class A:
    def __init__(self): ...


class B:
    def __init__(self, a: A):
        self.a = a


class Circular:
    def __init__(self, c: "Circular"):
        self.c = c


def test_resolver_simple_factory_dependency():
    container = Container()
    container.register_factory(A)

    resolver = Resolver(container)

    a = resolver.get((A, None))
    assert isinstance(a, A)


def test_resolver_instantiates_dependent():
    container = Container()
    container.register_factory(A)
    container.register_factory(B)

    resolver = Resolver(container)

    b = resolver.get((B, None))
    assert isinstance(b, B)
    assert isinstance(b.a, A)


@pytest.mark.xfail
def test_resolver_circular_dependency():
    container = Container()
    container.register_factory(Circular)

    resolver = Resolver(container)

    with pytest.raises(CircularDependencyError):
        resolver.get((Circular, None))


def test_resolver_fails_on_string_dependencies():
    container = Container()
    container.register_factory(Circular)

    resolver = Resolver(container)

    with pytest.raises(NotImplementedError) as e:
        resolver.get((Circular, None))

    assert str(e.value) == "String annotations are not supported yet"


@pytest.mark.parametrize("use_autowiring", [True, False])
def test_resolver_with_non_type_args(use_autowiring: bool):
    container = Container()
    resolver = Resolver(container)

    class A:
        def __init__(self, val: int): ...

    if not use_autowiring:
        container.register_factory(A)

    with pytest.raises(UnfillableArgumentError) as e:
        resolver.get((A, None))

    assert str(e.value) == "Unable to fill parameter val (<class 'int'>)"


def test_resolver_with_non_type_default_arg():
    container = Container()
    resolver = Resolver(container)

    class A:
        val: int

        def __init__(self, val: int = 1):
            self.val = val

    a = resolver.get((A, None))

    assert isinstance(a, A)
    assert a.val == 1
